import datetime
from typing import Any, Dict, List, Tuple, Union
from urllib.parse import urljoin

import requests


class Task:
    def __init__(self):
        self.task_id: str = ""
        self.owner: str = ""
        self.task_queue: str = ""
        self.payload: str = ""
        self.state: str = ""
        self.timeout: int = 0
        self.retries: int = 0
        self.max_retries: int = 0
        self.created_at: datetime.datetime = datetime.datetime.now()
        self.updated_at: datetime.datetime = datetime.datetime.now()
        self.result: Union[None, str] = None
        self.not_before: int = 0

    @classmethod
    def from_json(cls, payload: Dict[str, Any]) -> "Task":
        task = cls()

        task.task_id = payload["task_id"]
        task.owner = payload["owner"]
        task.task_queue = payload["task_queue"]
        task.payload = payload["payload"]
        task.state = payload["state"]
        task.timeout = payload["timeout"]
        task.retries = payload["retries"]
        task.max_retries = payload["max_retries"]
        task.created_at = datetime.datetime.fromtimestamp(payload["created_at"])
        task.updated_at = datetime.datetime.fromtimestamp(payload["updated_at"])

        if "result" in payload:
            task.result = payload["result"]

        task.not_before = payload["not_before"]

        return task


class Client:
    def __init__(self, maestro_endpoint: str):
        self.__maestro_endpoint = maestro_endpoint

    def launch_task(
        self,
        owner: str,
        queue: str,
        task_payload: str,
        retries: int = 0,
        timeout: int = 900,
        executes_in: int = 0,
        start_timeout: int = 0,
    ) -> str:
        """Launches a task.

        Launches a task to be executed by Maestro.

        Args:
            owner: owner of the task. Used for fair scheduling.
            queue: name of the queue to use.
            task_payload: encoded payload for the task.
            retries: Number of allowed retries in case of fail or timeouts.
            timeout: Allowed time span for the task to execute.
            executes_in: Number of seconds to wait before executing the task
            start_timeout: Allowed time span in seconds for the task to start.

        Returns:
            A string representing the Maestro task id

        Raises:
            ValueError: Problem in the communication with maestro.
        """

        resp = requests.post(
            urljoin(self.__maestro_endpoint, "/api/task/create"),
            json=self.__serialize_task(
                owner, queue, task_payload, retries, timeout, executes_in, start_timeout
            ),
        )
        if resp.status_code > 400 or "error" in resp.json():
            raise ValueError(
                f"Could not communicate with maestro. Status code is {resp.status_code}, "
                f"response is {resp.content}"
            )
        return resp.json()["task_id"]

    def task_state(self, task_id: str) -> Task:
        """Return the state of a task given its id.

        Args:
            task_id: the maestro id for this task

        Returns:
            A Task Object

        Raises:
            FileNotFoundError: This task does not exists
            ValueError: Error in communication with maestro

        """
        resp = requests.post(
            urljoin(self.__maestro_endpoint, "/api/task/get"),
            json={"task_id": task_id},
        )
        if resp.status_code == 404:
            raise FileNotFoundError("Could not find this task")
        elif resp.status_code > 400 or "error" in resp.json():
            raise ValueError(
                f"Could not communicate with maestro. Status code is {resp.status_code}, "
                f"response is {resp.content}"
            )

        return Task.from_json(resp.json()["task"])

    def next(self, queue: str) -> Union[Task, None]:
        """Get the following pending task.

        Get the following task to execute.

        Args:
            queue: the queue name

        Returns:
            A Task Object or None is no task is available

        Raises:
            ValueError: Error in communication with maestro
        """
        resp = requests.post(
            urljoin(self.__maestro_endpoint, "/api/queue/next"),
            json={"queue": queue},
        )

        if resp.status_code > 400 or "error" in resp.json():
            raise ValueError(
                f"Could not communicate with maestro. Status code is {resp.status_code}, "
                f"response is {resp.content}"
            )

        return Task.from_json(resp.json()["task"]) if resp.json() != {} else None

    def consume(self, queue: str) -> Union[Task, None]:
        """Consumes a task result from the queue.

        Consuming is getting the next result and directly removing it from
        maestro.

        Args:
            queue: the queue name

        Returns:
            A Task Object or None is no result is available

        Raises:
            ValueError: Error in communication with maestro

        """
        resp = requests.post(
            urljoin(self.__maestro_endpoint, "/api/queue/results/consume"),
            json={"queue": queue},
        )

        if resp.status_code > 400 or "error" in resp.json():
            raise ValueError(
                f"Could not communicate with maestro. Status code is {resp.status_code}, "
                f"response is {resp.content}"
            )

        return Task.from_json(resp.json()["task"]) if resp.json() != {} else None

    def delete_task(self, task_id: str) -> None:
        """Delete a task from maestro.

        Given its id the task is removed from maestro.

        Args:
            task_id: the identifier of the task

        Raises:
            FileNotFoundError: This task does not exists
            ValueError: Error in communication with maestro
        """
        resp = requests.post(
            urljoin(self.__maestro_endpoint, "/api/task/delete"),
            json={"task_id": task_id},
        )
        if resp.status_code == 404:
            raise FileNotFoundError("Could not find this task")
        elif resp.status_code > 400 or "error" in resp.json():
            raise ValueError(
                f"Could not communicate with maestro. Status code is {resp.status_code}, "
                f"response is {resp.content}"
            )

        return resp.json()

    def cancel_task(self, task_id: str) -> None:
        """Cancel a task from maestro.

        Given its id the task is cancelled from maestro.

        Args:
            task_id: the identifier of the task

        Raises:
            FileNotFoundError: This task does not exists
            ValueError: Error in communication with maestro
        """
        resp = requests.post(
            urljoin(self.__maestro_endpoint, "/api/task/cancel"),
            json={"task_id": task_id},
        )
        if resp.status_code == 404:
            raise FileNotFoundError("Could not find this task")
        elif resp.status_code > 400 or "error" in resp.json():
            raise ValueError(
                f"Could not communicate with maestro. Status code is {resp.status_code}, "
                f"response is {resp.content}"
            )

        return resp.json()

    def complete_task(self, task_id: str, result: str) -> None:
        """Complete a task in maestro.

        Given its id and its result the task is marked completed from maestro.

        Args:
            task_id: the identifier of the task
            result: the result for the task

        Raises:
            FileNotFoundError: This task does not exists
            ValueError: Error in communication with maestro
        """
        resp = requests.post(
            urljoin(self.__maestro_endpoint, "/api/task/complete"),
            json={"task_id": task_id, "result": result},
        )
        if resp.status_code == 404:
            raise FileNotFoundError("Could not find this task")
        elif resp.status_code > 400 or "error" in resp.json():
            raise ValueError(
                f"Could not communicate with maestro. Status code is {resp.status_code}, "
                f"response is {resp.content}"
            )

        return resp.json()

    def fail_task(self, task_id: str) -> None:
        """Fail a task in maestro.

        Given its id the task is marked failed in maestro.

        Args:
            task_id: the identifier of the task

        Raises:
            FileNotFoundError: This task does not exists
            ValueError: Error in communication with maestro
        """
        resp = requests.post(
            urljoin(self.__maestro_endpoint, "/api/task/fail"),
            json={"task_id": task_id},
        )
        if resp.status_code == 404:
            raise FileNotFoundError("Could not find this task")
        elif resp.status_code > 400 or "error" in resp.json():
            raise ValueError(
                f"Could not communicate with maestro. Status code is {resp.status_code}, "
                f"response is {resp.content}"
            )

        return resp.json()

    def launch_task_list(
        self,
        tasks: List[Tuple[str, str, str]],
        retries: int = 0,
        timeout: int = 900,
        executes_in: int = 0,
        start_timeout: int = 0,
    ) -> List[str]:
        """Launches a list a task.

        Each task is represented a 3-uple that will be added to maestro.

        Args:
            tasks:
                The list of tasks to be added. Each task is a 3-uple containing
                1. The owner of the task (as str)
                2. The queue
                3. The payload
            retries: Number of allowed retries in case of fail or timeouts.
            timeout: Allowed time span for the task to execute.
            executes_in: Number of seconds to wait before executing the task
            start_timeout: Allowed time span in seconds for the task to start.

        Returns:
            A list of string representing the identifiers of the tasks


        Raises:
            ValueError: Error in communication with maestro
        """
        payload = []
        for (owner, task_name, task_payload) in tasks:
            payload.append(
                self.__serialize_task(
                    owner,
                    task_name,
                    task_payload,
                    retries,
                    timeout,
                    executes_in,
                    start_timeout,
                )
            )

        resp = requests.post(
            urljoin(self.__maestro_endpoint, "/api/task/create/list"),
            json={"tasks": payload},
        )
        if resp.status_code > 400 or "error" in resp.json():
            raise ValueError(
                f"Could not communicate with maestro. Status code is {resp.status_code}, "
                f"response is {resp.content}"
            )
        return resp.json()["task_ids"]

    @staticmethod
    def __serialize_task(
        owner: str,
        queue: str,
        task_payload: str,
        retries: int,
        timeout: int,
        executes_in: int,
        start_timeout: int,
    ) -> dict[str, Any]:
        task = {
            "owner": owner,
            "queue": queue,
            "retries": retries,
            "timeout": timeout,
            "payload": task_payload,
        }

        if executes_in > 0:
            task["not_before"] = datetime.datetime.now().timestamp() + executes_in

        if start_timeout > 0:
            task["startTimeout"] = datetime.datetime.now().timestamp() + executes_in

        return task



from abc import ABC, abstractmethod


class Cache(ABC):
    @abstractmethod
    def get(self, key: str) -> str:
        ...

    @abstractmethod
    def put(self, key: str, value: str, ttl: int):
        ...

    @abstractmethod
    def delete(self, key: str):
        ...

    @abstractmethod
    def set_ttl(self, key: str, ttl: int):
        ...



from typing import List, Tuple
from uuid import uuid4

from maestro_python_client.Cache.Cache import Cache
from maestro_python_client.Client import Client, Task


class CachedClient(Client):
    def __init__(
        self,
        maestro_endpoint: str,
        cache: Cache,
        cached_queues: list[str] = [],
        completed_task_ttl: int = 900,
        **kwargs,
    ) -> None:
        super().__init__(maestro_endpoint, **kwargs)
        self.__cached_queues: set[str] = set(cached_queues)
        self.__cache = cache
        self.__completed_task_ttl = completed_task_ttl

    def launch_task(
        self,
        owner: str,
        queue: str,
        task_payload: str,
        retries: int = 0,
        timeout: int = 900,
        executes_in: int = 0,
        start_timeout: int = 0,
    ) -> str:
        """Launches a task.

        Launches a task to be executed by Maestro.

        Args:
            owner: owner of the task. Used for fair scheduling.
            queue: name of the queue to use.
            task_payload: encoded payload for the task.
            retries: Number of allowed retries in case of fail or timeouts.
            timeout: Allowed time span for the task to execute.
            executes_in: Number of seconds to wait before executing the task
            start_timeout: Allowed time span in seconds for the task to start. Must be > 0 if the task is in a cached queue.

        Returns:
            A string representing the Maestro task id

        Raises:
            ValueError: Problem in the communication with maestro or invalid start_time.
        """

        if queue in self.__cached_queues and start_timeout <= 0:
            raise ValueError("Start timeout must be > 0 for cached task")

        ttl = (
            executes_in
            + self.__completed_task_ttl
            + (start_timeout + timeout) * (retries + 1)
        )
        task_payload = self.__cache_payload(
            queue,
            task_payload,
            ttl,
        )
        return super().launch_task(
            owner, queue, task_payload, retries, timeout, executes_in, start_timeout
        )

    def next(self, queue: str) -> Task | None:
        task = super().next(queue)
        return self.__task_from_cache(task)

    def consume(self, queue: str) -> Task | None:
        task = super().consume(queue)
        return self.__task_from_cache(task)

    def task_state(self, task_id: str) -> Task:
        task = super().task_state(task_id)
        return self.__task_from_cache(task)  # type: ignore

    def complete_task(self, task_id: str, result: str) -> None:
        task = super().task_state(task_id)

        self.__set_ttl(task.task_queue, task.payload, self.__completed_task_ttl)
        result = self.__cache_payload(
            task.task_queue, result, self.__completed_task_ttl
        )

        super().complete_task(task_id, result)

    def cancel_task(self, task_id: str) -> None:
        task = super().task_state(task_id)
        self.__set_ttl(task.task_queue, task.payload, self.__completed_task_ttl)

        super().cancel_task(task_id)

    def fail_task(self, task_id: str) -> None:
        task = super().task_state(task_id)
        self.__set_ttl(task.task_queue, task.payload, self.__completed_task_ttl)

        super().fail_task(task_id)

    def delete_task(self, task_id: str) -> None:
        task = super().task_state(task_id)
        self.__delete(task.task_queue, task.payload)
        if task.result:
            self.__delete(task.task_queue, task.result)

        super().delete_task(task_id)

    def launch_task_list(
        self,
        tasks: List[Tuple[str, str, str]],
        retries: int = 0,
        timeout: int = 900,
        executes_in: int = 0,
        start_timeout: int = 0,
    ) -> List[str]:
        """Launches a list a task.

        Each task is represented a 3-uple that will be added to maestro.

        Args:
            tasks:
                The list of tasks to be added. Each task is a 3-uple containing
                1. The owner of the task (as str)
                2. The queue
                3. The payload
            retries: Number of allowed retries in case of fail or timeouts.
            timeout: Allowed time span for the task to execute.
            executes_in: Number of seconds to wait before executing the task
            start_timeout: Allowed time span in seconds for the task to start. Must be > 0 if the task is in a cached queue.

        Returns:
            A list of string representing the identifiers of the tasks


        Raises:
            ValueError: Error in communication with maestro or invalid start_time
        """

        payload_ttl = (
            executes_in
            + self.__completed_task_ttl
            + (start_timeout + timeout) * (retries + 1)
        )

        for i, (owner, queue, payload) in enumerate(tasks):
            if queue in self.__cached_queues and start_timeout <= 0:
                raise ValueError("Start timeout must be > 0 for cached task")

            tasks[i] = (
                owner,
                queue,
                self.__cache_payload(
                    queue,
                    payload,
                    payload_ttl,
                ),
            )

        return super().launch_task_list(
            tasks, retries, timeout, executes_in, start_timeout
        )

    def __task_from_cache(self, task: Task | None) -> Task | None:
        if not task:
            return task

        task.payload = self.__payload_from_cache(task.task_queue, task.payload)

        if task.result:
            task.result = self.__payload_from_cache(task.task_queue, task.result)

        return task

    def __cache_payload(self, queue: str, payload: str, timeout: int = 0) -> str:
        if queue not in self.__cached_queues:
            return payload

        cache_key = self.__unique_key(queue)
        self.__cache.put(cache_key, payload, timeout)
        return cache_key

    def __payload_from_cache(self, queue: str, payload: str) -> str:
        if queue not in self.__cached_queues:
            return payload

        return self.__cache.get(payload)

    def __set_ttl(self, queue: str, key: str, ttl: int):
        if queue not in self.__cached_queues:
            return

        self.__cache.set_ttl(key, ttl)

    def __delete(self, queue: str, key: str):
        if queue not in self.__cached_queues:
            return

        self.__cache.delete(key)

    @staticmethod
    def __unique_key(queue_name: str) -> str:
        return f"maestro-cache-{queue_name}-{str(uuid4())}"



from collections import defaultdict
from copy import deepcopy
from unittest.mock import MagicMock, patch

from pytest import fixture, raises

from maestro_python_client.Cache.Cache import Cache
from maestro_python_client.CachedClient import CachedClient
from test_utils.String import unique_str


@fixture
def json_task():
    return {
        "task_id": unique_str(),
        "owner": unique_str(),
        "task_queue": unique_str(),
        "payload": "",
        "result": "",
        "state": "pending",
        "timeout": 300,
        "retries": 0,
        "max_retries": 0,
        "created_at": 1,
        "updated_at": 1,
        "not_before": 0,
    }


class TestCache(Cache):
    def __init__(self) -> None:
        super().__init__()
        self.__cache = {}

    @property
    def cache(self) -> dict[str, str]:
        return deepcopy(self.__cache)

    def put(self, key: str, value: str, ttl: int | None = None):
        self.__cache[key] = value

    def get(self, key: str) -> str:
        return self.__cache[key]

    def set_ttl(self, key: str, ttl: int):
        assert key in self.__cache

    def delete(self, key: str):
        self.__cache.pop(key)

    def key_for_value(self, value: str) -> str | None:
        for key, v in self.__cache.items():
            if v == value:
                return key

        return None


@patch("requests.post")
def test_launch_task(requests, subtests):
    response_mock = MagicMock()
    requests.return_value = response_mock
    response_mock.status_code = 200

    with subtests.test("Queue not cached"):
        mock = MagicMock()
        client = CachedClient("", mock, ["cached"])
        client.launch_task(unique_str(), "not_cached", "payload", start_timeout=100)
        assert not mock.put.called

    with subtests.test("Queue cached"):
        mock = MagicMock()
        client = CachedClient("", mock, ["cached"])
        client.launch_task(unique_str(), "cached", "payload", start_timeout=100)
        assert mock.put.called

    with subtests.test("Start timeout 0 cached should fail"):
        mock = MagicMock()
        client = CachedClient("", mock, ["cached"])
        with raises(ValueError):
            client.launch_task(unique_str(), "cached", "payload", start_timeout=0)

    with subtests.test("Start timeout 0 not cached should not fail"):
        mock = MagicMock()
        client = CachedClient("", mock, ["cached"])
        client.launch_task(unique_str(), "not_cached", "payload", start_timeout=0)


@patch("requests.post")
def test_launch_task_list(requests, subtests):
    response_mock = MagicMock()
    requests.return_value = response_mock
    response_mock.status_code = 200

    with subtests.test("Queue not cached"):
        mock = MagicMock()
        client = CachedClient("", mock, ["cached"])
        client.launch_task_list(
            [(unique_str(), "not_cached", "payload")], start_timeout=100
        )
        assert not mock.put.called

    with subtests.test("Queue cached"):
        mock = MagicMock()
        client = CachedClient("", mock, ["cached"])
        client.launch_task_list(
            [(unique_str(), "cached", "payload")], start_timeout=100
        )
        assert mock.put.called

    with subtests.test("Start timeout 0 cached should fail"):
        mock = MagicMock()
        client = CachedClient("", mock, ["cached"])
        with raises(ValueError):
            client.launch_task_list(
                [(unique_str(), "cached", "payload")], start_timeout=0
            )

    with subtests.test("Start timeout 0 not cached should not fail"):
        mock = MagicMock()
        client = CachedClient("", mock, ["cached"])
        client.launch_task_list(
            [(unique_str(), "not_cached", "payload")], start_timeout=0
        )


@patch("requests.post")
def test_task_getters(requests, subtests, json_task):
    response_mock = MagicMock()
    requests.return_value = response_mock
    response_mock.status_code = 200

    methods = ["task_state", "next", "consume"]
    for method in methods:
        with subtests.test("With cache"):
            cache = TestCache()
            client = CachedClient("", cache, ["cached"])

            response_mock.json.return_value = defaultdict(lambda: "")
            client.launch_task(unique_str(), "cached", "payload", start_timeout=100)

            response_mock.json.return_value = {
                "task": {
                    **json_task,
                    "task_queue": "cached",
                    "payload": cache.key_for_value("payload"),
                }
            }
            client.complete_task("task", "result")
            assert len(cache.cache) == 2

            response_mock.json.return_value = {
                "task": {
                    **json_task,
                    "task_queue": "cached",
                    "result": cache.key_for_value("result"),
                    "payload": cache.key_for_value("payload"),
                }
            }
            test_method = getattr(client, method)
            res = test_method("cached")
            assert res
            assert res.payload == "payload"
            assert res.result == "result"

        with subtests.test("Without cache"):
            cache = TestCache()
            client = CachedClient("", cache, ["cached"])

            response_mock.json.return_value = defaultdict(lambda: "")
            client.launch_task(unique_str(), "not_cached", "payload", start_timeout=100)

            response_mock.json.return_value = {
                "task": {
                    **json_task,
                    "task_queue": "not_cached",
                    "payload": "payload",
                }
            }
            client.complete_task("task", "result")
            assert not cache.cache

            response_mock.json.return_value = {
                "task": {
                    **json_task,
                    "task_queue": "not_cached",
                    "result": "result",
                    "payload": "payload",
                }
            }

            test_method = getattr(client, method)
            res = test_method("not_cached")
            assert res
            assert res.payload == "payload"
            assert res.result == "result"


@patch("requests.post")
def test_complete_task(requests, subtests, json_task):
    response_mock = MagicMock()
    requests.return_value = response_mock
    response_mock.status_code = 200

    with subtests.test("With cache"):
        cache = TestCache()
        client = CachedClient("", cache, ["cached"])

        response_mock.json.return_value = defaultdict(lambda: "")
        client.launch_task(unique_str(), "cached", "payload", start_timeout=100)

        response_mock.json.return_value = {
            "task": {
                **json_task,
                "task_queue": "cached",
                "payload": cache.key_for_value("payload"),
            }
        }
        client.complete_task("task", "result")
        assert cache.key_for_value("result")

    with subtests.test("Without cache"):
        cache = TestCache()
        client = CachedClient("", cache, ["cached"])

        response_mock.json.return_value = defaultdict(lambda: "")
        client.launch_task(unique_str(), "not_cached", "payload", start_timeout=100)

        response_mock.json.return_value = {
            "task": {
                **json_task,
                "task_queue": "not_cached",
                "payload": cache.key_for_value("payload"),
            }
        }
        client.complete_task("task", "result")
        assert not cache.key_for_value("result")


@patch("requests.post")
def test_delete_task(requests, subtests, json_task):
    response_mock = MagicMock()
    requests.return_value = response_mock
    response_mock.status_code = 200

    with subtests.test("With cache and result"):
        cache = TestCache()
        client = CachedClient("", cache, ["cached"])

        response_mock.json.return_value = defaultdict(lambda: "")
        client.launch_task(unique_str(), "cached", "payload", start_timeout=100)
        response_mock.json.return_value = {
            "task": {
                **json_task,
                "task_queue": "cached",
                "payload": cache.key_for_value("payload"),
            }
        }
        client.complete_task("task", "result")
        assert cache.key_for_value("result")
        assert cache.key_for_value("payload")

        response_mock.json.return_value = {
            "task": {
                **json_task,
                "task_queue": "cached",
                "payload": cache.key_for_value("payload"),
                "result": cache.key_for_value("result"),
            }
        }
        client.delete_task("task")
        assert not cache.key_for_value("result")
        assert not cache.key_for_value("payload")

    with subtests.test("With cache no result"):
        cache = TestCache()
        client = CachedClient("", cache, ["cached"])

        response_mock.json.return_value = defaultdict(lambda: "")
        client.launch_task(unique_str(), "cached", "payload", start_timeout=100)

        response_mock.json.return_value = {
            "task": {
                **json_task,
                "task_queue": "cached",
                "payload": cache.key_for_value("payload"),
            }
        }
        assert not cache.key_for_value("result")
        assert cache.key_for_value("payload")

        client.delete_task("task")
        assert not cache.key_for_value("result")
        assert not cache.key_for_value("payload")

    with subtests.test("Without cache"):
        cache = TestCache()
        client = CachedClient("", cache, ["cached"])

        response_mock.json.return_value = defaultdict(lambda: "")
        client.launch_task(unique_str(), "not_cached", "payload", start_timeout=100)

        response_mock.json.return_value = {
            "task": {
                **json_task,
                "task_queue": "not_cached",
                "payload": cache.key_for_value("payload"),
            }
        }
        client.complete_task("task", "result")
        assert not cache.key_for_value("result")
        assert not cache.key_for_value("payload")

        client.delete_task("task")
        assert not cache.key_for_value("result")
        assert not cache.key_for_value("payload")


@patch("requests.post")
def test_task_stoppers(requests, subtests, json_task):
    response_mock = MagicMock()
    requests.return_value = response_mock
    response_mock.status_code = 200

    methods = ["fail_task", "cancel_task"]
    for method in methods:
        with subtests.test("With cache"):
            cache = MagicMock()
            client = CachedClient("", cache, ["cached"])

            response_mock.json.return_value = defaultdict(lambda: "")
            client.launch_task(unique_str(), "cached", "payload", start_timeout=100)
            response_mock.json.return_value = {
                "task": {
                    **json_task,
                    "task_queue": "cached",
                    "result": cache.key_for_value("result"),
                    "payload": cache.key_for_value("payload"),
                }
            }
            test_method = getattr(client, method)
            test_method("cached")
            assert cache.set_ttl.called

        with subtests.test("Without cache"):
            cache = MagicMock()
            client = CachedClient("", cache, ["cached"])

            response_mock.json.return_value = defaultdict(lambda: "")
            client.launch_task(unique_str(), "not_cached", "payload", start_timeout=100)

            response_mock.json.return_value = {
                "task": {
                    **json_task,
                    "task_queue": "not_cached",
                    "result": "result",
                    "payload": "payload",
                }
            }

            test_method = getattr(client, method)
            test_method("not_cached")
            assert not cache.set_ttl.called



from redis import Redis

from maestro_python_client.Cache.Cache import Cache


class RedisCache(Cache):
    def __init__(self, redis: Redis) -> None:
        super().__init__()
        self.__redis = redis

    def get(self, key: str) -> str:
        payload = self.__redis.get(key)

        if not payload:
            raise ValueError(f"Key {key} is not cached")

        return payload.decode("utf-8")

    def put(self, key: str, value: str, ttl: int | None = None):
        if not self.__redis.set(key, value.encode("utf-8"), ex=ttl):
            raise ValueError(f"Could not add {key} to cache")

    def delete(self, key: str):
        self.__redis.delete(key)

    def set_ttl(self, key: str, ttl: int):
        self.__redis.expire(key, ttl)



from pytest import raises

from maestro_python_client.Cache.RedisCache import RedisCache
from test_utils.Redis import new_test_redis
from test_utils.String import unique_str


def test_get(subtests):
    cache = RedisCache(new_test_redis())

    with subtests.test("key exists"):
        key = unique_str()
        cache.put(key, "value")

        assert cache.get(key) == "value"

    with subtests.test("get twice"):
        key = unique_str()
        cache.put(key, "value")

        assert cache.get(key) == "value"

    with subtests.test("value doesn't exists"):
        with raises(ValueError):
            cache.get(unique_str())


def test_put(subtests):
    cache = RedisCache(new_test_redis())

    with subtests.test("put new"):
        key = unique_str()
        cache.put(key, "value")

        assert cache.get(key) == "value"

    with subtests.test("put existing"):
        key = unique_str()
        cache.put(key, "value")
        cache.put(key, "new")

        assert cache.get(key) == "new"


def test_delete(subtests):
    cache = RedisCache(new_test_redis())

    with subtests.test("put new"):
        key = unique_str()
        cache.put(key, "value")
        cache.delete(key)

        with raises(ValueError):
            cache.get(key)


def test_set_ttl(subtests):
    redis = new_test_redis()
    cache = RedisCache(redis)

    with subtests.test("put new"):
        key = unique_str()
        cache.put(key, "value")
        cache.set_ttl(key, 42)

        assert redis.ttl(key) == 42
